-------------------------------------------------------------------------------
--                G  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

-- Account & Character data saved for the Niggles addon
NigglesDB  = {};
NigglesDBC = {};

-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local OPTIONS_OFFSET_LEFT    = 16;
local OPTIONS_OFFSET_TOP     = -16;
local OPTIONS_OFFSET_HEADING = 8;
local OPTIONS_OFFSET_SPACING = 5;

local NEW_FEATURE_ICON =
  "|TInterface\\OptionsFrame\\UI-OptionsFrame-NewFeatureIcon:0:0:0:0|t";

-- Required Account and Character data saved for the Niggles addon
local requiredDB =
{
  ["raidAssists"   ] = 1,
  ["characterRoles"] = 1,
};
local requiredDBC =
{
  ["achievements"  ] = 1,
  ["equipSetSpec"  ] = 1,
  ["faction"       ] = 1,
  ["mounts"        ] = 1,
  ["pets"          ] = 1,
  ["settings"      ] = 1
};

-- Default values for all Niggles Character settings
local defaultDBC =
{
  -- General Options
  ["UiAchievementProgress"] = true,
  ["UiAchievementSearch"]   = false,
  ["UiAutoLootArch"]        = false,
  ["UiAutoLootCurrency"]    = false,
  ["UiAutoLootFish"]        = false,
  ["UiAutoLootMoney"]       = false,
  ["UiChatClassColor"]      = true,
  ["UiCompanionsFiltered"]  = true,
  ["UiEMIgnoreToggle"]      = true,
  ["UiEMSetSpec"]           = true,
  ["UiFirstLastPage"]       = true,
  ["UiFlightList"]          = true,
  ["UiFlightListShown"]     = true,
  ["UiGarrisonMission"]     = true,
  ["UiGuildInviteAll"]      = false,
  ["UiGuildRosterFilter"]   = true,
  ["UiItemDelete"]          = true,
  ["UiLowLevelItems"]       = false,
  ["UiMailFont"]            = true,
  ["UiMapLocation"]         = true,
  ["UiMerchantBuyMax"]      = true,
  ["UiMerchantSearch"]      = true,
  ["UiMiniMapZoom"]         = true,
  ["UiPetBandage"]          = true,
  ["UiPetJournalTooltip"]   = true,
  ["UiPetsOwned"]           = true,
  ["UiQuestRewards"]        = true,

  -- Warning Options
  ["WarningCurrency"]       = true,
  ["WarningCurrencyMask"]   = 0x02,
  ["WarningEquipSetItem"]   = true,
  ["WarningRod"]            = true,
  ["WarningLance"]          = true,
  ["WarningLowLevelItems"]  = true,
  ["WarningNoForm"]         = true,
  ["WarningPetBattle"]      = true,
  ["WarningPetGrowl"]       = false,
  ["WarningTradeSkillCap"]  = true,
  ["LogChat"]               = false,
  ["LogCombat"]             = false,

  -- Profession Options
  ["ProfGeneralHistory"]    = true,
  ["ProfGeneralReagents"]   = true,
  ["ProfArchHilite"]        = true,
  ["ProfArchKeystones"]     = true,
  ["ProfArchEnhIcon"]       = true,
  ["ProfArchRClickBack"]    = true,
  ["ProfArchProgress"]      = true,

  -- Raids Options
  ["RaidsAutoPromote"]      = true,
  ["RaidsAutoLootMethod"]   = true,
  ["RaidsLootMethod"]       = "master",
  ["RaidsAutoLootThold"]    = true,
  ["RaidsLootThreshold"]    = 4,

  -- Calendar Event Options
  ["CalEventsList"]         = true,
  ["CalEventsInvites"]      = true,
  ["CalEventsOnline"]       = true,
  ["CalEventsRoles"]        = true,
  ["CalEventsRoleFilter"]   = 0x8001,

  -- Companion Options
  ["MountsEnabled"]         = true,
  ["MountsClassRace"]       = true,
  ["MountsNagrandTalbuk"]   = true,
  ["MountsHoliday"]         = false,
  ["PetsEnabled"]           = true,
  ["PetsInAreas"]           = NIGGLES_PETAREAS_DEFAULT,

  -- Locator Options
  ["LocatorShown"]          = false,
  ["LocatorX"]              = -1,
  ["LocatorY"]              = -1,
};

local optionsLayouts =
{
  ["General"] =
  {
    -- UI Enhancements
    {
      objType    = "heading",
      text       = "UiEnhancementsText",
      subText    = "UiEnhancementsSubText",
    },
    {
      objType    = "checkbutton",
      setting    = "UiAchievementProgress",
    },
    {
      objType    = "checkbutton",
      setting    = "UiCompanionsFiltered",
      newFeature = true,
    },
    {
      objType    = "checkbutton",
      setting    = "UiChatClassColor",
    },
    {
      objType    = "checkbutton",
      setting    = "UiEMIgnoreToggle",
    },
    {
      objType    = "checkbutton",
      setting    = "UiEMSetSpec",
    },
    {
      objType    = "checkbutton",
      setting    = "UiFirstLastPage",
    },
    {
      objType    = "checkbutton",
      setting    = "UiGarrisonMission",
      newFeature = true,
    },
    {
      objType    = "checkbutton",
      setting    = "UiGuildInviteAll",
    },
    {
      objType    = "checkbutton",
      setting    = "UiLowLevelItems",
    },
    {
      objType    = "checkbutton",
      setting    = "UiMailFont",
      newColumn  = 1,
    },
    {
      objType    = "checkbutton",
      setting    = "UiMapLocation",
    },
    {
      objType    = "checkbutton",
      setting    = "UiMerchantBuyMax",
    },
    {
      objType    = "checkbutton",
      setting    = "UiMiniMapZoom",
    },
    {
      objType    = "checkbutton",
      setting    = "UiPetBandage",
      newFeature = true,
    },
    {
      objType    = "checkbutton",
      setting    = "UiPetJournalTooltip",
    },
    {
      objType    = "checkbutton",
      setting    = "UiPetsOwned",
    },
    {
      objType    = "checkbutton",
      setting    = "UiQuestRewards",
    },
    {
      objType    = "checkbutton",
      setting    = "UiItemDelete",
    },
    -- Search & Filtering
    {
      objType    = "heading",
      text       = "UiSearchFilterText",
      subText    = "UiSearchFilterSubText",
    },
    {
      objType    = "checkbutton",
      setting    = "UiAchievementSearch",
    },
    {
      objType    = "checkbutton",
      setting    = "UiFlightList",
    },
    {
      objType    = "checkbutton",
      setting    = "UiGuildRosterFilter",
      newColumn  = 1,
    },
    {
      objType    = "checkbutton",
      setting    = "UiMerchantSearch",
    },
    -- Auto-Loot
    {
      objType    = "heading",
      text       = "UiAutoLootText",
      subText    = "UiAutoLootSubText",
    },
    {
      objType    = "checkbutton",
      setting    = "UiAutoLootArch",
    },
    {
      objType    = "checkbutton",
      setting    = "UiAutoLootCurrency",
    },
    {
      objType    = "checkbutton",
      setting    = "UiAutoLootFish",
      newColumn  = 1,
    },
    {
      objType    = "checkbutton",
      setting    = "UiAutoLootMoney",
    },
  },
  ["Warnings"] =
  {
    -- Warnings
    {
      objType    = "heading",
      text       = "WarningsText",
      subText    = "WarningsSubText",
    },
    {
      objType    = "checkbutton",
      setting    = "WarningCurrency",
      dependants = 1,
    },
    {
      objType    = "dropdown",
      setting    = "WarningCurrencyMask",
      dependency = "WarningCurrency",
      width      = 120,
      xOffset    = 26,
      items      =
      {
        {
          type    = "radioButton",
          label   = L["Weekly"],
          value   = 0x01,
        },
        {
          type    = "radioButton",
          label   = L["Total"],
          value   = 0x02,
        },
        {
          type    = "radioButton",
          label   = L["WeeklyTotal"],
          value   = 0x03,
        },
      },
    },
    {
      objType    = "checkbutton",
      setting    = "WarningEquipSetItem",
    },
    {
      objType    = "checkbutton",
      setting    = "WarningRod",
    },
    {
      objType    = "checkbutton",
      setting    = "WarningLance",
    },
    {
      objType    = "checkbutton",
      setting    = "WarningLowLevelItems",
      newColumn  = 1,
    },
    {
      objType    = "checkbutton",
      setting    = "WarningNoForm",
      init       = "NigglesOptionsWarningsInit",
    },
    {
      objType    = "checkbutton",
      setting    = "WarningPetBattle",
    },
    {
      objType    = "checkbutton",
      setting    = "WarningPetGrowl",
    },
    {
      objType    = "checkbutton",
      setting    = "WarningTradeSkillCap",
    },
    -- Logs
    {
      objType    = "heading",
      text       = "LogsText",
      subText    = "LogsSubText",
    },
    {
      objType    = "checkbutton",
      setting    = "LogChat",
    },
    {
      objType    = "checkbutton",
      setting    = "LogCombat",
      newColumn  = 1,
    },
  },
  ["Professions"] =
  {
    -- General
    {
      objType    = "heading",
      text       = "ProfGeneralText",
      subText    = "ProfGeneralSubText",
    },
    {
      objType    = "checkbutton",
      setting    = "ProfGeneralHistory",
    },
    {
      objType    = "checkbutton",
      setting    = "ProfGeneralReagents",
      newColumn  = 1,
    },
    -- Archaeology
    {
      objType    = "heading",
      text       = "ProfArchaeologyText",
      subText    = "ProfArchaeologySubText",
    },
    {
      objType    = "checkbutton",
      setting    = "ProfArchProgress",
    },
    {
      objType    = "checkbutton",
      setting    = "ProfArchEnhIcon",
    },
    {
      objType    = "checkbutton",
      setting    = "ProfArchHilite",
    },
    {
      objType    = "checkbutton",
      setting    = "ProfArchKeystones",
      newColumn  = 1,
    },
    {
      objType    = "checkbutton",
      setting    = "ProfArchRClickBack",
    },
  },
  ["Raids"] =
  {
    {
      objType    = "heading",
      text       = "RaidsText",
      subText    = "RaidsSubText",
    },
    {
      objType    = "checkbutton",
      setting    = "RaidsAutoLootMethod",
      dependants = 1,
    },
    {
      objType    = "dropdown",
      setting    = "RaidsLootMethod",
      dependency = "RaidsAutoLootMethod",
      adjacent   = 1,
      xOffset    = 5,
      width      = 135,
      items      =
      {
        {
          type    = "radioButton",
          label = L["FreeForAll"],
          value = "freeforall"
        },
        {
          type    = "radioButton",
          label = L["RoundRobin"],
          value = "roundrobin"
        },
        {
          type    = "radioButton",
          label = L["MasterLooter"],
          value = "master"
        },
        {
          type    = "radioButton",
          label = L["GroupLoot"],
          value = "group"
        },
        {
          type    = "radioButton",
          label = L["NeedGreed"],
          value = "needbeforegreed"
        }
      },
    },
    {
      objType    = "checkbutton",
      setting    = "RaidsAutoLootThold",
      dependants = 1,
    },
    {
      objType    = "dropdown",
      setting    = "RaidsLootThreshold",
      dependency = "RaidsAutoLootThold",
      adjacent   = 1,
      width      = 90,
      xOffset    = 5,
      items      =
      {
        {
          type    = "radioButton",
          label = ITEM_QUALITY2_DESC,
          value = 2,
          color = ITEM_QUALITY_COLORS[2].hex
        },
        {
          type    = "radioButton",
          label = ITEM_QUALITY3_DESC,
          value = 3,
          color = ITEM_QUALITY_COLORS[3].hex
        },
        {
          type    = "radioButton",
          label = ITEM_QUALITY4_DESC,
          value = 4,
          color = ITEM_QUALITY_COLORS[4].hex
        }
      },
    },
    {
      objType    = "checkbutton",
      setting    = "RaidsAutoPromote",
      dependants = 1,
    },
    {
      objType    = "custom",
      name       = "NigglesOptionsRaidsAssists",
      xOffset    = 26,
      init       = "NigglesOptionsRaidsInit",
      dependency = "RaidsAutoPromote",
    },
  },
  -- Calendar Events
  ["CalEvents"] =
  {
    {
      objType    = "heading",
      text       = "CalEventsText",
      subText    = "CalEventsSubText",
    },
    {
      objType    = "checkbutton",
      setting    = "CalEventsList",
    },
    {
      objType    = "checkbutton",
      setting    = "CalEventsInvites",
    },
    {
      objType    = "checkbutton",
      setting    = "CalEventsOnline",
    },
    {
      objType    = "checkbutton",
      setting    = "CalEventsRoles",
      dependants = 1,
    },
    {
      objType    = "custom",
      name       = "NigglesOptionsCalEventsCharacterRoles",
      xOffset    = 26,
      init       = "NigglesOptionsCalEventsInit",
      dependency = "CalEventsRoles",
    },
  },
  -- Companions
  ["Companions"] =
  {
    -- Mounts
    {
      objType    = "heading",
      text       = "MountsText",
      subText    = "MountsSubText",
      setting    = "MountsEnabled",
      dependants = 1,
    },
    {
      objType    = "checkbutton",
      setting    = "MountsClassRace",
      dependants = 1,
      dependency = NigglesOptionsMountsClassRaceEnable,
    },
    {
      objType    = "custom",
      name       = "NigglesOptionsCompanionsClassRaceButton1",
      dependency = NigglesOptionsMountsClassRaceEnable,
      xOffset    = 26,
    },
    {
      objType    = "checkbutton",
      setting    = "MountsNagrandTalbuk",
      dependency = "MountsEnabled",
      newFeature = true,
    },
    {
      objType    = "checkbutton",
      setting    = "MountsHoliday",
      dependency = "MountsEnabled",
    },
    {
      objType    = "custom",
      name       = "NigglesOptionsCompanionsMountsClear",
      newLine    = 1,
      dependency = "MountsEnabled",
    },
    {
      objType    = "custom",
      name       = "NigglesOptionsCompanionsMountsSelect",
      adjacent   = 1,
      dependency = "MountsEnabled",
    },
    -- Pets
    {
      objType    = "heading",
      text       = "PetsText",
      subText    = "PetsSubText",
      setting    = "PetsEnabled",
      dependants = 1,
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInArea"],
      tooltip    = L["PetsInAreaTooltip"],
      value      = GetContinentName(7), -- Draenor
      mask       = NigglesPetsGetAreaFlag,
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInArea"],
      tooltip    = L["PetsInAreaTooltip"],
      value      = GetContinentName(2), -- Eastern Kingdoms
      mask       = NigglesPetsGetAreaFlag,
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInArea"],
      tooltip    = L["PetsInAreaTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = GetContinentName(1), -- Kalimdor
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInArea"],
      tooltip    = L["PetsInAreaTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = GetContinentName(4), -- Northrend
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInArea"],
      tooltip    = L["PetsInAreaTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = GetContinentName(3), -- Outland
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInArea"],
      tooltip    = L["PetsInAreaTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = GetContinentName(6), -- Pandaria
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInCities"],
      tooltip    = L["PetsInCitiesTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = "city",
      dependency = "PetsEnabled",
      newColumn  = 1,
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInScenarios"],
      tooltip    = L["PetsInScenariosTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = "scenario",
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInDungeons"],
      tooltip    = L["PetsInDungeonsTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = "party",
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInRaids"],
      tooltip    = L["PetsInRaidsTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = "raid",
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInArenas"],
      tooltip    = L["PetsInArenasTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = "arena",
      dependency = "PetsEnabled",
    },
    {
      objType    = "checkbutton",
      setting    = "PetsInAreas",
      label      = L["PetsInBG"],
      tooltip    = L["PetsInBGTooltip"],
      mask       = NigglesPetsGetAreaFlag,
      value      = "pvp",
      dependency = "PetsEnabled",
    },
    {
      objType    = "custom",
      name       = "NigglesOptionsCompanionsPetsClear",
      newLine    = 1,
      dependency = "PetsEnabled",
    },
    {
      objType    = "custom",
      name       = "NigglesOptionsCompanionsPetsSelect",
      adjacent   = 1,
      dependency = "PetsEnabled",
    },
  },
};

local optionsObjects =
{
  checkButtons = {},
  numCheckBtns = 0,
  custom       = {},
  numCustom    = 0,
  dropDowns    = {},
  numDropDowns = 0,
  headings     = {},
  numHeadings  = 0,
};

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local optionsDropDownItemOnCheck;
local optionsDropDownItemOnClick;
local optionsDropDownOnClick;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to update the value of a dropdown when one of its options is
-- clicked.
--
optionsDropDownItemOnCheck = function(menu, menuArgs, index, value)
  -- Local Variables
  local anchor = menu:GetAnchor();

  return (value == Niggles.updatedOptions.settings[anchor.nglSetting]);
end

--
-- Function to update the value of a dropdown when one of its options is
-- clicked.
--
optionsDropDownItemOnClick = function(menu, menuArgs, index, value)
  -- Local Variables
  local items = menu:GetItems();
  local anchor = menu:GetAnchor();

  -- Save the new value
  Niggles.updatedOptions.settings[anchor.nglSetting] = value;

  -- Set the drop down menu's value
  anchor:SetText((items[index].color or "")..items[index].label);

  return;
end

--
-- Function to display a menu for a dropdown when it is clicked on
--
optionsDropDownOnClick = function(self, button)
  -- Display the menu for the dropdown
  Niggles.menu:Toggle(self.menuData, self);

  return;
end

--
-- Function called when a Niggles Options heading is checked/unchecked
--
local function optionsHeadingSetChecked(self, checked)
  -- Call the original 'SetChecked' function
  self.origSetChecked(self, checked);

  -- Update the state of the sub-text
  NigglesOptionsHeadingOnEnable(self)

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to cancel changes to the options
--
function NigglesOptionsCancel()
  -- Destroy any updated options
  Niggles.updatedOptions = nil;

  return;
end

--
-- Function to process check button clicks in the Niggles Options frame
--
function NigglesOptionsCheckButtonOnClick(self, mouseButton)
  -- Local Variables
  local settings = Niggles.updatedOptions.settings;

  -- Update the check box's setting
  if (self:GetChecked()) then
    PlaySound("igMainMenuOptionCheckBoxOn");
    if (self.nglMask ~= nil) then
      settings[self.nglSetting] =
        bit.bor(settings[self.nglSetting], self.nglMask);
    else
      settings[self.nglSetting] = true;
    end
  else
    PlaySound("igMainMenuOptionCheckBoxOff");
    if (self.nglMask ~= nil) then
      settings[self.nglSetting] =
        bit.band(settings[self.nglSetting], bit.bnot(self.nglMask));
    else
      settings[self.nglSetting] = false;
    end
  end

  -- Enable any dependants, if required
  if (self.nglDependants == 1) then
    NigglesOptionsEnable(self:GetParent());
  end

  return;
end

--
-- Function to set the tooltip for a widget in the Niggles Options panels
--
function NigglesOptionsCheckButtonOnEnter(self)
  -- Check if a tooltip was found
  if (self.nglTooltip ~= nil) then
    GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT", 20);
    GameTooltip:SetText(self.nglTooltip, nil, nil, nil, nil, 1);
    GameTooltip:Show();
  end

  return;
end

--
-- Function to initialise an Niggles Options check button
--
function NigglesOptionsCheckButtonOnLoad(self)
  -- Local Variables
  local fontString;

  -- Change the points for the check button's textures to allow resizing
  -- Note: This assumes that the check button has only the standard textures.
  for _, region in pairs({self:GetRegions()}) do
    if (region:GetObjectType() == "Texture") then
      region:ClearAllPoints();
      region:SetPoint("TOPLEFT");
      region:SetWidth(26);
      region:SetHeight(26);
    end
  end

  -- Set the check button's text to force creation of the font string
  self:SetText(" ");

  -- Adjust the position of the font string
  fontString = self:GetFontString();
  fontString:SetPoint("TOP", 0, -6);
  fontString:SetPoint("LEFT", 26, 0);

  return;
end

--
-- Function to enable/disable objects in Niggles Options panels
--
function NigglesOptionsEnable(self)
  -- Local Variables
  local settings = Niggles.updatedOptions.settings;
  local currentObj;

  -- Enable/disable check buttons, based on any dependency
  for idx = 1, optionsObjects.numCheckBtns do
    currentObj = optionsObjects.checkButtons[idx];
    if ((type(currentObj.nglDependency) == "string") and
        (settings[currentObj.nglDependency] ~= nil)) then
      currentObj:SetEnabled(settings[currentObj.nglDependency]);
    elseif (type(currentObj.nglDependency) == "function") then
      currentObj.nglDependency(currentObj, currentObj.nglSetting);
    end
  end

  -- Enable/disable custom object, based on any dependency
  for idx = 1, optionsObjects.numCustom do
    currentObj = optionsObjects.custom[idx];
    if ((type(currentObj.nglDependency) == "string") and
        (settings[currentObj.nglDependency] ~= nil) and
        (type(currentObj.SetEnabled) == "function")) then
      currentObj:SetEnabled(settings[currentObj.nglDependency]);
    elseif (type(currentObj.nglDependency) == "function") then
      currentObj.nglDependency(currentObj, currentObj.nglSetting);
    end
  end

  -- Enable/disable drop down menus, based on any dependency
  for idx = 1, optionsObjects.numDropDowns do
    currentObj = optionsObjects.dropDowns[idx];
    if ((type(currentObj.nglDependency) == "string") and
        (settings[currentObj.nglDependency] ~= nil)) then
      currentObj:SetEnabled(settings[currentObj.nglDependency]);
    elseif (type(currentObj.nglDependency) == "function") then
      currentObj.nglDependency(currentObj, currentObj.nglSetting);
    end
  end
  HideDropDownMenu(1);

  -- Call the panel's enable function, if required
  if (type(self.nglEnable) == "function") then
    self.nglEnable(self);
  end

  return;
end

--
-- Function called when a Niggles Options heading is created
--
function NigglesOptionsHeadingOnLoad(self)
  -- Perform the same initialisation as a Niggles Options check button
  NigglesOptionsCheckButtonOnLoad(self);

  -- Adjust the position of the font string
  fontString = self:GetFontString();
  fontString:SetPoint("TOP", 0, -5);

  -- Override the check heading's 'SetChecked' method
  self.origSetChecked = self.SetChecked;
  self.SetChecked     = optionsHeadingSetChecked;

  return;
end

--
-- Function to change all the Niggles settings to their default values
--
function NigglesOptionsDefaults(frame)
  -- Destroy any updated options
  Niggles.updatedOptions = nil;

  -- Set all the settings to their default values
  for key, value in pairs(defaultDBC) do
    NigglesDBC.settings[key] = value;
  end

  -- Clear the character lists
  NigglesDB.raidAssists    = {};
  NigglesDB.characterRoles = {};

  -- Clear all the companions
  NigglesDBC.mounts = {};
  NigglesDBC.pets   = {};

  -- Re-initialise the settings
  NigglesOptionsRefresh();

  return;
end

--
-- Function called when a Niggles Options Heading is clicked
--
function NigglesOptionsHeadingOnClick(self, mouseButton)
  -- Check the heading has a setting
  if (self.nglSetting ~= nil) then
    -- Perform the same actions as a Niggles Options check button
    NigglesOptionsCheckButtonOnClick(self, mouseButton)

    -- Update the state of the sub-text
    NigglesOptionsHeadingOnEnable(self)
  end

  return;
end

--
-- Function called when a Niggles Options heading is enabled/disabled
--
function NigglesOptionsHeadingOnEnable(self)
  -- Update the state of the sub-text
  if ((self.nglSetting == nil) or
      ((self:IsEnabled()) and (self:GetChecked()))) then
    self.subText:SetFontObject("GameFontHighlightSmall");
  else
    self.subText:SetFontObject("GameFontDisableSmall");
  end

  return;
end

--
-- Function called when an Options panel is shown. It lays out the objects for
-- the Options panel and calls any custom function specified.
--
function NigglesOptionsPanelOnShow(self)
  -- Local Variables
  local fontString;
  local indent = OPTIONS_OFFSET_LEFT;
  local layout;
  local numCheckBtns = 0;
  local numCustom = 0;
  local numDropDowns = 0;
  local numHeadings = 0;
  local numObjects = 0;
  local object;
  local objType;
  local prevObject;
  local prevYPos;
  local settings;
  local textWidth;
  local xPos = OPTIONS_OFFSET_LEFT
  local yPos = OPTIONS_OFFSET_TOP;
  local yPosHeading = yPos;
  local yPosMin = yPos;

  -- Check if the option panel should be laid out
  if ((self == nil) or
      (self.nglLayout == nil) or
      (optionsLayouts[self.nglLayout] == nil) or
      (not self:IsShown()) or
      (not self:IsVisible()) or
      (not InterfaceOptionsFramePanelContainer:IsVisible()) or
      (Niggles.updatedOptions == nil)) then
    return;
  end

  -- Assign some info to more convenient variables
  settings = Niggles.updatedOptions.settings;

  -- Lay out objects for the options panel
  for idx, layout in ipairs(optionsLayouts[self.nglLayout]) do
    -- Initialise some variables
    object = nil;
    objType = layout.objType:lower();

    -- Lay out the next object
    if (objType == "checkbutton") then
      -- Increment the number of check buttons
      numCheckBtns = numCheckBtns+1;

      -- Create a new check button, if required
      if (numCheckBtns > #optionsObjects.checkButtons) then
        optionsObjects.checkButtons[numCheckBtns] = CreateFrame("checkbutton",
          nil, self, "NigglesOptionsCheckButtonTemplate");
      end
      object = optionsObjects.checkButtons[numCheckBtns];

      -- Set up the check button
      object.nglSetting = layout.setting;
      object.nglValue   = layout.value;
      if (type(layout.mask) == "function") then
        object.nglMask = layout.mask(layout.value);
      else
        object.nglMask = layout.mask;
      end
      if (layout.value ~= nil) then
        if (layout.newFeature) then
          object:SetFormattedText(
            (layout.label or L[layout.setting])..NEW_FEATURE_ICON,
            layout.value);
        else
          object:SetFormattedText((layout.label or L[layout.setting]),
            layout.value);
        end
        object.nglTooltip = format(
          (layout.tooltip or L[layout.setting.."Tooltip"]), layout.value);
      else
        if (layout.newFeature) then
          object:SetText((layout.label or L[layout.setting])..NEW_FEATURE_ICON);
        else
          object:SetText(layout.label or L[layout.setting]);
        end
        object.nglTooltip = (layout.tooltip or L[layout.setting.."Tooltip"]);
      end
      object:Enable();
      if (object.nglMask ~= nil) then
        object:SetChecked(
          bit.band((settings[layout.setting] or 0), object.nglMask) ~= 0);
      else
        object:SetChecked(settings[layout.setting] or false);
      end
    elseif (objType == "heading") then
      -- Increment the number of headings
      numHeadings = numHeadings+1;

      -- Create a new heading, if required
      if (numHeadings > #optionsObjects.headings) then
        optionsObjects.headings[numHeadings] = CreateFrame("checkbutton",
          nil, self, "NigglesOptionsHeadingTemplate");
      end
      object = optionsObjects.headings[numHeadings];

      -- Enable/disable user input
      object:EnableKeyboard(layout.setting ~= nil);
      object:EnableMouse(layout.setting ~= nil);

      -- Set up the heading
      object.nglSetting = layout.setting;
      object.nglTooltip = layout.tooltipText;
      object:SetText(L[layout.text]);
      object.subText:SetText(L[layout.subText]);
      object:GetNormalTexture():Show();
      object:Enable();
      object:SetChecked(settings[layout.setting] or false);

      -- Set up the heading's textures and font strings
      for _, region in pairs({object:GetRegions()}) do
        if (region:GetObjectType() == "Texture") then
          if (layout.setting == nil) then
            region:Hide();
          end
        elseif (region:GetObjectType() == "FontString") then
          region:SetPoint("LEFT", (layout.setting ~= nil and 26 or 0), 0);
        end
      end
    elseif (objType == "dropdown") then
      -- Increment the number of dropdown
      numDropDowns = numDropDowns+1;

      -- Create a new dropdown, if required
      if (numDropDowns > #optionsObjects.dropDowns) then
        object = CreateFrame("Button", nil, self, "NigglesDropDownTemplate");
        object:SetScript("OnClick", optionsDropDownOnClick);
        object.menuData =
        {
          style    = "menu",
          position = "bottom",
          offset   = {x = -2, y = 2},
          onClick  = optionsDropDownItemOnClick,
          onCheck  = optionsDropDownItemOnCheck,
          items    = nil
        };
        optionsObjects.dropDowns[numDropDowns] = object;
      else
        object = optionsObjects.dropDowns[numDropDowns];
      end

      -- Set up the drop down menu
      object.nglSetting = layout.setting;
      object.menuData.items = layout.items;
      object:SetWidth(layout.width or 100);
      object:Enable();

      -- Set the dropdown's value
      setting = settings[layout.setting]
      for _, current in ipairs(layout.items) do
        if (current.value == setting) then
          object:SetText((current.color or "")..current.label);
          break;
        end
      end
    elseif (objType == "custom") then
      -- Get the custom object
      object = _G[layout.name];

      -- Add the object to the list of custom objects
      numCustom = numCustom+1;
      optionsObjects.custom[numCustom] = object;
    end

    -- Check there is an object to lay out
    if (object ~= nil) then
      -- Save some settings
      object.nglDependants = layout.dependants;
      object.nglDependency = layout.dependency;

      -- Adjust the current position, based on layout options
      if (objType == "heading") then
        indent = OPTIONS_OFFSET_LEFT;
        xPos   = indent;
        yPos   = yPosMin-(numObjects > 0 and OPTIONS_OFFSET_HEADING or 0);
      elseif ((objType ~= "heading") and
              (layout.newColumn == 1)) then
        -- Adjust the current position
        xPos = xPos+200;
        yPos = yPosHeading;
      elseif (layout.newLine == 1) then
        xPos = indent;
      end

      -- Perform any custom initialisation
      if ((layout.init ~= nil) and (type(_G[layout.init]) == "function")) then
        _G[layout.init](object, layout.setting);
      end

      -- Perform any post-processing for the object
      if (objType == "checkbutton") then
        textWidth = object:GetFontString():GetStringWidth();
        if ((textWidth ~= nil) and (textWidth > 0)) then
          object:SetWidth(26+textWidth);
        end
      end

      -- Position and show the object
      if ((objType ~= "heading") and
          (layout.adjacent == 1) and
          (prevObject ~= nil)) then
        object:SetPoint("TOPLEFT", prevObject, "TOPRIGHT",
          (layout.xOffset or 0), (layout.yOffset or 0));
      else
        object:SetPoint("TOPLEFT", xPos+(layout.xOffset or 0),
          yPos+(layout.yOffset or 0));
        yPos = yPos-(layout.height or
          object:GetHeight()+OPTIONS_OFFSET_SPACING);
      end
      if (objType ~= "custom") then
        object:SetParent(self);
        object:Show(true);
      end

      -- Update the current position
      if ((objType == "heading") and (layout.setting ~= nil)) then
        indent = OPTIONS_OFFSET_LEFT+26;
        xPos   = indent;
      end
      yPosMin = math.min(yPosMin, yPos);
      if (objType == "heading") then
        yPosHeading = yPos;
      end

      -- Save the object
      prevObject = object;

      -- Increment the total number of objects
      numObjects = numObjects+1;
    end
  end

  -- Hide any unused objects
  for idx = numCheckBtns+1, #optionsObjects.checkButtons do
    optionsObjects.checkButtons[idx]:Hide();
  end
  for idx = numDropDowns+1, #optionsObjects.dropDowns do
    optionsObjects.dropDowns[idx]:Hide();
  end
  for idx = numHeadings+1, #optionsObjects.headings do
    optionsObjects.headings[idx]:Hide();
  end

  -- Save the number of objects of each type
  optionsObjects.numCheckBtns = numCheckBtns;
  optionsObjects.numCustom    = numCustom;
  optionsObjects.numDropDowns = numDropDowns;
  optionsObjects.numHeadings  = numHeadings;

  -- Enable/disable objects based on current settings
  NigglesOptionsEnable(self);

  return;
end

--
-- Function to refresh the options for the Niggles Options panels
--
function NigglesOptionsRefresh()
  -- Check there aren't any updated options already
  if ((InterfaceOptionsFramePanelContainer:IsVisible()) and
      (Niggles.updatedOptions == nil)) then
    -- Initialise the updated options
    Niggles.updatedOptions = {};

    -- Copy the existing settings
    Niggles.updatedOptions.settings = CopyTable(NigglesDBC.settings);

    -- Copy the existing character lists
    Niggles.updatedOptions.raidAssists    = CopyTable(NigglesDB.raidAssists);
    Niggles.updatedOptions.characterRoles = CopyTable(NigglesDB.characterRoles);

    -- Copy the existing lists of companions
    Niggles.updatedOptions.mounts = CopyTable(NigglesDBC.mounts);
    Niggles.updatedOptions.pets   = CopyTable(NigglesDBC.pets);
  end

  return;
end

--
-- Function to save updated options for the Niggles Options panels
--
function NigglesOptionsSave()
  -- Check if there are any updated options
  if (Niggles.updatedOptions ~= nil) then
    -- Save the new settings
    NigglesDBC.settings = Niggles.updatedOptions.settings;

    -- Save the new lists of companions
    NigglesDBC.mounts = Niggles.updatedOptions.mounts;
    NigglesDBC.pets   = Niggles.updatedOptions.pets;

    -- Save the new character lists
    NigglesDB.raidAssists    = Niggles.updatedOptions.raidAssists;
    NigglesDB.characterRoles = Niggles.updatedOptions.characterRoles;

    -- Register for events
    NigglesRegisterForEvents();

    -- Remove the updated options
    Niggles.updatedOptions = nil;

    -- Force any visible calendar event frame to update itself
    if ((CalendarCreateEventFrame) and
        (CalendarCreateEventFrame:IsShown())) then
      NigglesCalenderInvitesShow(CalendarCreateEventFrame);
    end
    if ((CalendarViewEventFrame) and
        (CalendarViewEventFrame:IsShown())) then
      NigglesCalenderInvitesShow(CalendarViewEventFrame);
    end

    -- Show/Hide the most valuable quest reward, if required
    if ((QuestFrame:IsShown()) or
        (QuestLogFrame:IsShown()) or
        (QuestLogDetailFrame:IsShown())) then
      NigglesQuestsItemInfoUpdate();
    end

    -- Force the Archaeology pages to update themselves, if required
    if ((ArchaeologyFrameSummaryPage) and
        (ArchaeologyFrameSummaryPage:IsShown())) then
      NigglesArchaeologySummaryUpdate();
    end
    if ((ArchaeologyFrameCompletedPage) and
        (ArchaeologyFrameCompletedPage:IsShown())) then
      NigglesArchaeologyCompletedUpdate();
    end

    -- Set up the toggles in the Mount & Pet Journal
    NigglesCompanionsTogglesSetup();
  end

  return;
end

--
-- Function to perform custom initialisation for objects in the Options
-- Warnings panel
--
function NigglesOptionsWarningsInit(object, setting)
  -- Perform any custom initialisation for the object specified
  if (setting == "WarningNoForm") then
    local formName = (Niggles.playerForm or L["Form"]);
    object:SetFormattedText(L["WarningNoForm"], formName);
    object.nglTooltip = format(L["WarningNoFormTooltip"],
      string.lower(formName));
    if (Niggles.playerForm == nil) then
      object:Disable();
    end
  end

  return;
end

--
-- Function to make sure all the settings are initialised
--
function NigglesSettingsInit()
  -- Destroy any obsolete data
  if (next(NigglesDB)) then
    for key, _ in pairs(NigglesDB) do
      if (requiredDB[key] == nil) then
        NigglesDB[key] = nil;
      end
    end
  end
  if (next(NigglesDBC)) then
    for key, _ in pairs(NigglesDBC) do
      if (requiredDBC[key] == nil) then
        NigglesDBC[key] = nil;
      end
    end
  end
  if ((NigglesDBC.settings ~= nil) and
      (next(NigglesDBC.settings))) then
    for key, _ in pairs(NigglesDBC.settings) do
      if (defaultDBC[key] == nil) then
        NigglesDBC.settings[key] = nil;
      end
    end
  end
  if ((NigglesDBC.mounts ~= nil) and
      (next(NigglesDBC.mounts))) then
    for key, value in pairs(NigglesDBC.mounts) do
      if (value == 0) then
         NigglesDBC.mounts[key] = nil;
      end
    end
  end

  -- Initialise any missing required Account or Character data
  for key, _ in pairs(requiredDB) do
    if (NigglesDB[key] == nil) then
      NigglesDB[key] = {};
    end
  end
  for key, _ in pairs(requiredDBC) do
    if (NigglesDBC[key] == nil) then
      NigglesDBC[key] = {};
    end
  end

  -- Make sure all settings are initialised
  for key, value in pairs(defaultDBC) do
    if (NigglesDBC.settings[key] == nil) then
      NigglesDBC.settings[key] = value;
    end
  end

  -- Adjust settings based on class/race
  if (Niggles.playerForm == nil) then
    NigglesDBC.settings.WarningNoForm = false;
  end

  return;
end
